# Creates a task-bar icon.  Run from Python.exe to see the
# messages printed.
from win32api import *
from win32gui import *
import win32con
import win32gui
import pydoc, threading, os, sys
import pydoc_win32

WM_USER_TASKBAR = win32con.WM_USER + 1
WM_USER_FIND_MODULE = win32con.WM_USER + 2
WM_USER_STOP_SERVER = win32con.WM_USER + 3
WM_USER_OPEN_BROWSER = win32con.WM_USER + 4

def openbrowser(url):
    try:
        import webbrowser
        webbrowser.open(url)
    except ImportError: # pre-webbrowser.py compatibility
        if sys.platform == 'win32':
            os.system('start "%s"' % url)
        elif sys.platform == 'mac':
            try:
                import ic
                ic.launchurl(url)
            except ImportError: pass
        else:
            rc = os.system('netscape -remote "openURL(%s)" &' % url)
            if rc: os.system('netscape "%s" &' % url)

class MainWindow:
        def __init__(self, url):
                self.url = url
                message_map = {
                        win32con.WM_DESTROY: self.OnDestroy,
                        win32con.WM_COMMAND: self.OnCommand,
                        WM_USER_TASKBAR : self.OnTaskbarNotify,
                }
                # Register the Window class.
                wc = WNDCLASS()
                hinst = wc.hInstance = GetModuleHandle(None)
                wc.lpszClassName = "Python Module Documentation"
                wc.style = win32con.CS_VREDRAW | win32con.CS_HREDRAW;
                wc.hCursor = LoadCursor( 0, win32con.IDC_ARROW )
                wc.hbrBackground = win32con.COLOR_WINDOW
                wc.lpfnWndProc = message_map # could also specify a wndproc.
                classAtom = RegisterClass(wc)
                # Create the Window.
                style = win32con.WS_OVERLAPPED | win32con.WS_SYSMENU
                self.hwnd = CreateWindow( classAtom, "Taskbar Demo", style, \
                        0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT, \
                        win32gui.GetDesktopWindow(), 0, hinst, None)
                UpdateWindow(self.hwnd)
                openbrowser(self.url)

             # Try and find a custom icon
                iconPathName = os.path.abspath(
                    os.path.join( os.path.split(sys.executable)[0],"as_pydoc" ,"pythonview.ico" ))

                if os.path.isfile(iconPathName):
                     icon_flags = win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE
                     hicon = LoadImage(hinst, iconPathName, win32con.IMAGE_ICON, 0, 0,
                                icon_flags)
                else:
                     print "Can't find a Python icon file - using default"
                     hicon = LoadIcon(0, win32con.IDI_APPLICATION)

                flags = NIF_ICON | NIF_MESSAGE | NIF_TIP
                self.nid = (self.hwnd, 0, flags, WM_USER_TASKBAR, hicon, "Python Module Documentation")
                Shell_NotifyIcon(NIM_ADD, self.nid)

        def OnDestroy(self, hwnd, msg, wparam, lparam):
                nid = (self.hwnd, 0)
                Shell_NotifyIcon(NIM_DELETE, nid)
                PostQuitMessage(0) # Terminate the app.

        def OnTaskbarNotify(self, hwnd, msg, wparam, lparam):
                if lparam in (win32con.WM_LBUTTONUP, win32con.WM_RBUTTONUP):
                        menu = CreatePopupMenu()
                        AppendMenu( menu, win32con.MF_STRING, WM_USER_OPEN_BROWSER,
                                    "&Open Browser")
                        AppendMenu( menu, win32con.MF_STRING, WM_USER_FIND_MODULE,
                                        "&Find Module Documentation")
                        AppendMenu( menu, win32con.MF_STRING, WM_USER_STOP_SERVER,
                                        "&Stop Documentation Server" )
                        pos = GetCursorPos()
                        TrackPopupMenu(menu, win32con.TPM_LEFTALIGN, 
                                                    pos[0], pos[1], 0, self.hwnd, None)
                return 1

        def OnCommand(self, hwnd, msg, wparam, lparam):
                try:
                        id = LOWORD(wparam)
                        if id == WM_USER_FIND_MODULE:
##                                mypath, _ = os.path.split(sys.argv[0])
##                                executablepath, _ = os.path.split(sys.executable)
##                                pythonw = os.path.join(executablepath, "pythonw.exe")
##                                filename = os.path.join(mypath, "astgui.py")
###                                print os.spawnl(os.P_NOWAIT, pythonw, filename, str(self.port))
###                                os.system(" ".join(["start", pythonw, filename, str(self.port)]))
##                                os.spawnl(os.P_NOWAIT, pythonw, pythonw, filename, str(self.port))
###                                import astgui
###                                astgui.gui("http://localhost:%d/" % self.port)
                                w=pydoc_win32.SearchWindow(self.url)
                                w.DoModal()
                        elif id == WM_USER_STOP_SERVER:
                                PostQuitMessage(0)
                                Shell_NotifyIcon(NIM_DELETE, self.nid)
                        elif id == WM_USER_OPEN_BROWSER:
                                openbrowser(self.url)
                        else:
                                print "Unknown command -", id
                except:
                        import traceback
                        traceback.print_exc()
                return 1

def ready(*args):
        pass

def main():
	host = (sys.platform == 'mac') and '127.0.0.1' or 'localhost'	
        port = 1750
        thread = threading.Thread(target=pydoc.serve, args=(port, ready))
        thread.setDaemon(1)
        thread.start()
        w=MainWindow("http://%s:%d/" % (host,port))
        PumpMessages()

if __name__=='__main__':
        main()
